/*
* Binaeruhr
* Autor: Karsten Mueller
* email@kmkcl.de
*/

#ifndef F_CPU
        #define F_CPU 8000000UL
#endif

#include "USI_TWI_Master.h"

#include <avr/interrupt.h>
#include  <avr/io.h>
#include <stdlib.h>
#include <stdio.h>
#include <util/delay.h>
#include "binuhr.h"
#include <string.h>

ISR (TIMER0_OVF_vect) { //Multiplex-Timer
  static uint8_t kanal, trigger=0;
  static uint8_t ct0, ct1, rpt;
  uint8_t i, leds;

  i=*anzeige.helligkeit;
  if(i>0||trigger%5==0) {       //für "super-dunkel" Modus, nur jedes 5. Mal Anzeige an
    leds=anzeige.leds[kanal];
    DDRB=ausgaenge[kanal];
    DDRD=leds;
    PORTD=leds|(1<<PD2);
    
    /* für "super-dunkel" und "dunkel": Anzeige gleich wieder ausschalten.
     * PWM Interrupt würde zu spät kommen.
     * */
    if(i<=1) {                  
      PORTD=(1<<PD2);
      DDRD=0;
      DDRB=0x5F;
    }
    
    if(++kanal>5) kanal=0;
  }
  /*
   * Tastenentprellung nach Peter Dannegger 
   * danni@specs.de
   * siehe auch: http://www.mikrocontroller.net/articles/Entprellung
   */
  
  if(++trigger==40) {
    trigger=0;
    i = key_state ^ ~PINA;              // key changed ?
    ct0 = ~( ct0 & i );                 // reset or count ct0
    ct1 = ct0 ^ (ct1 & i);              // reset or count ct1
    i &= ct0 & ct1;                     // count until roll over ?
    key_state ^= i;                     // then toggle debounced state
    key_press |= key_state & i;         // 0->1: key press detect

    if( (key_state & REPEAT_MASK) == 0 )        // check repeat function
      rpt = REPEAT_START;               // start delay
    if( --rpt == 0 ) {
      rpt = REPEAT_NEXT;                        // repeat delay
      key_rpt |= key_state & REPEAT_MASK;
    }
  }
}

ISR (TIMER0_COMPA_vect) { //Helligkeiten: Soft-PWM innerhalb Multiplex
  PORTD=(1<<PD2);
  DDRD=0;    
  DDRB=0x5F;
}

ISR(INT0_vect) { //Sekundentakt von DS1307
  zeit.ss++;
}

uint8_t get_key_short( uint8_t key_mask ) {
  cli();
  key_mask=~key_state & key_mask;       // read key state and key press atomic !
  key_mask &= key_press;                // read key(s)
  key_press ^= key_mask;  
  sei();
  return key_mask;
}

uint8_t get_key_long( uint8_t key_mask ) {
  cli();
  key_mask &= key_rpt;                  // read key(s)
  key_rpt ^= key_mask;                  // clear key(s)
  key_mask &= key_press;                // read key(s)
  key_press ^= key_mask;   
  sei();
  return key_mask;
}

void anzeigen(uint8_t a,uint8_t b,uint8_t c) {
  uint8_t i=0;
  for(i=0;i<6;i++) {
    anzeige.leds[i]=((a&1)<<PD4) | ((b&1)<<PD6) | ((c&1)<<PD5); // 6,5,4 meine Binäruhr
    a>>=1;
    b>>=1;
    c>>=1;
  }
}

char bcd2int(char bcd) {
  return (bcd&0x0F)+10*(bcd>>4);
}

void programTime(void) {
  cli();
  memset(daten, 0, 18);
  /* Array mit 0 initialisieren, somit wird:
   * daten[2]=0 (DS1307: Oszillator einschalten, Sekunden auf 0 setzen)
   * daten[16]=0 (DS3231: Ausgabe Sekundentakt auf SQW)
   * daten[17]=0 (DS3231: 32kHz abschalten)
   */
  daten[0]=DS1307;
  daten[3]=(zeit.mm%10) | ((zeit.mm/10)<<4);
  daten[4]=(zeit.hh%10) | ((zeit.hh/10)<<4);
  daten[9]=16; //DS1307: Ausgabe Sekundentakt 
  USI_TWI_Master_Initialise();
  USI_TWI_Start_Read_Write(daten,18);
  sei();
}

/*Hauptprogramm*/
uint8_t modus, ss_alt, temp;

int main(void)
{
  PORTA=(1<<PA0) | (1<<PA1);  //interner Pullup Taster                                                                                                                                                                                                                          //interner Pullup für Taster
  PORTD=(1<<PD2);               //interner Pullup für INT0
  TCCR0B=(1<<CS01);     //Multiplex: Timer0  1/8 CPU-Takt
  TIMSK=(1<<TOIE0) | (1<<OCIE0A);       //Overflow Interrupt und Compare-Match Interrupt
  MCUCR=3;              //Interrupt INT0 bei steigender Taktflanke
  GIMSK=(1<<INT0) ;     //Interrupt INT0 aktivieren
        
  /*DS1307 ansprechen- aktuelle Uhrzeit auslesen
  * und einschalten, wenn außer Betrieb (zum Beispiel nach Batteriewechsel)
  */
  USI_TWI_Master_Initialise();
  daten[0]=DS1307;
  daten[1]=0;
  USI_TWI_Start_Random_Read( daten , 5 );
  temp=daten[1];
  zeit.mm=daten[2];
  zeit.hh=daten[3];

  if(temp&128)   //DS1307 Oszillator war ausgeschaltet
    programTime();
  
  //ausgelesenes BCD in Binärzahlen umwandeln
  zeit.ss=bcd2int(temp);
  zeit.mm=bcd2int(zeit.mm);
  zeit.hh=bcd2int(zeit.hh);
  temp=0;
  modus=0;
  anzeige.helligkeit=((zeit.hh>=TAG)&&(zeit.hh<NACHT))?&hell:&dunkel;
  
  sei(); //Interrupts einschalten, Anzeige aktiv
  
  while(1) {
    
    OCR0A=stufen[*(anzeige.helligkeit)]; //Helligkeit setzen
    
    if(modus==0) { // Uhrzeit anzeigen
      anzeigen(zeit.hh, zeit.ss, zeit.mm);
      
      if(zeit.ss>59){
        zeit.ss-=60; //Sekunden werden während "Helligkeit" verändern weiter gezählt
        zeit.mm++;
      }
      if(zeit.mm>59){
        zeit.mm=0;
        zeit.hh++;
        if(zeit.hh==TAG) anzeige.helligkeit=&hell;
        if(zeit.hh==NACHT) anzeige.helligkeit=&dunkel;
      }
      if(zeit.hh>23) zeit.hh=0;
      
      if(get_key_short(1<<KEY_UP)) { //Helligkeit ändern
        modus=1;
        ss_alt=zeit.ss;
      }
      if(get_key_short(1<<KEY_DOWN)) //manueller TAG-NACHT Wechsel
        anzeige.helligkeit=(anzeige.helligkeit==&hell)?&dunkel:&hell;
      
      if(get_key_long(1<<KEY_DOWN)) //Uhr stellen
        modus=2;
    }
    
    if(modus==1) { //Helligkeit verändern
      anzeigen((anzeige.helligkeit==&dunkel)?1:2,*(anzeige.helligkeit)+1,*(anzeige.helligkeit)+1);
      if(get_key_short(1<<KEY_UP)) {
        *anzeige.helligkeit=*(anzeige.helligkeit)+1;
        ss_alt=zeit.ss;
      }
      if(get_key_short(1<<KEY_DOWN)) {
        *anzeige.helligkeit=*(anzeige.helligkeit)-1;
        ss_alt=zeit.ss;
      }
      if(*anzeige.helligkeit>128) *anzeige.helligkeit=0; //negativ
      if(*anzeige.helligkeit==stufenLength) *anzeige.helligkeit=stufenLength-1;
      if(zeit.ss>ss_alt+2) modus=0; //nach ca. 1-2 Sekunden ohne Eingabe zurückspringen
    }
    
    if(modus==2) { //Stunden stellen
      anzeige.helligkeit=&hell;
      anzeigen(zeit.hh,0,0);

      if(get_key_short(1<<KEY_UP))  {
        zeit.hh++; 
        if(zeit.hh>23) zeit.hh=0;
      }
      
      if(get_key_short(1<<KEY_DOWN)) {
        zeit.hh--;
        if(zeit.hh>128) zeit.hh=23;
      }
      
      if(get_key_long(1<<KEY_DOWN))
        modus++;
    }
    
    if(modus==3) { //Minuten stellen
      anzeigen(0, 0, zeit.mm);
      if(get_key_short(1<<KEY_UP))  {
        zeit.mm++; 
        if(zeit.mm>59) zeit.hh=0;
      }
      if(get_key_short(1<<KEY_DOWN)) {
        zeit.mm--;
        if(zeit.mm>128) zeit.mm=59;
      }
      if(get_key_long(1<<KEY_DOWN)) {
        zeit.ss=0;
        programTime();
        modus=0;
      }
    }
  }
}
